import pytest

from great_expectations.core.expectation_diagnostics.expectation_diagnostics import (
    ExpectationDiagnostics,
)
from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationLegacyTestCaseAdapter,
    ExpectationTestDataCases,
    TestData,
)
from great_expectations.core.expectation_diagnostics.supporting_types import (
    AugmentedLibraryMetadata,
    ExpectationDescriptionDiagnostics,
    ExpectationDiagnosticCheckMessage,
    ExpectationDiagnosticMaturityMessages,
    ExpectationRendererDiagnostics,
    ExpectationTestDiagnostics,
    RendererTestDiagnostics,
)

# module level markers
pytestmark = pytest.mark.unit

expectation_test_data_case = ExpectationTestDataCases(
    dataset_name="expectation_diagnostics_test_data",
    data=TestData(
        **{
            "a": ["aaa", "abb", "acc", "add", "bee"],
            "b": ["aaa", "abb", "acc", "bdd", None],
            "column_name with space": ["aaa", "abb", "acc", "add", "bee"],
        }
    ),
    tests=[
        ExpectationLegacyTestCaseAdapter(
            **{
                "title": "negative_test_insufficient_mostly_and_one_non_matching_value",
                "exact_match_out": False,
                "in": {"column": "a", "regex": "^a", "mostly": 0.9},
                "out": {
                    "success": False,
                    "unexpected_index_list": [4],
                    "unexpected_list": ["bee"],
                },
                "suppress_test_for": ["sqlite", "mssql"],
            }
        ),
        ExpectationLegacyTestCaseAdapter(
            **{
                "title": "positive_test_exact_mostly_w_one_non_matching_value",
                "exact_match_out": False,
                "in": {"column": "a", "regex": "^a", "mostly": 0.8},
                "out": {
                    "success": True,
                    "unexpected_index_list": [4],
                    "unexpected_list": ["bee"],
                },
                "suppress_test_for": ["sqlite", "mssql"],
            }
        ),
    ],
)

edr = ExpectationDiagnostics(
    description=ExpectationDescriptionDiagnostics(
        **{
            "camel_name": "ExpectColumnValuesToMatchRegex",
            "snake_name": "expect_column_values_to_match_regex",
            "short_description": "Expect column entries to be strings that match a given regular expression.",  # noqa: E501 # FIXME CoP
            "docstring": 'Expect column entries to be strings that match a given regular expression.\n    \n    Valid matches can be found     anywhere in the string, for example "[at]+" will identify the following strings as expected: "cat", "hat",     "aa", "a", and "t", and the following strings as unexpected: "fish", "dog".\n\n    expect_column_values_to_match_regex is a     :func:`column_map_expectation <great_expectations.execution_engine.execution_engine.MetaExecutionEngine\n    .column_map_expectation>`.\n\n    Args:\n        column (str):             The column name.\n        regex (str):             The regular expression the column entries should match.\n\n    Keyword Args:\n        mostly (None or a float between 0 and 1):             Return `"success": True` if at least mostly fraction of values match the expectation.             For more detail, see :ref:`mostly`.\n\n    Other Parameters:\n        result_format (str or None):             Which output mode to use: `BOOLEAN_ONLY`, `BASIC`, `COMPLETE`, or `SUMMARY`.\n            For more detail, see :ref:`result_format <result_format>`.\n        catch_exceptions (boolean or None):             If True, then catch exceptions and include them as part of the result object.             For more detail, see :ref:`catch_exceptions`.\n        meta (dict or None):             A JSON-serializable dictionary (nesting allowed) that will be included in the output without             modification. For more detail, see :ref:`meta`.\n\n    Returns:\n        An ExpectationSuiteValidationResult\n\n        Exact fields vary depending on the values passed to :ref:`result_format <result_format>` and\n        :ref:`catch_exceptions`, and :ref:`meta`.\n\n    See Also:\n        :func:`expect_column_values_to_not_match_regex         <great_expectations.execution_engine.execution_engine.ExecutionEngine\n        .expect_column_values_to_not_match_regex>`\n\n        :func:`expect_column_values_to_match_regex_list         <great_expectations.execution_engine.execution_engine.ExecutionEngine\n        .expect_column_values_to_match_regex_list>`\n\n    ',  # noqa: E501 # FIXME CoP
        }
    ),
    library_metadata=AugmentedLibraryMetadata(
        **{
            "maturity": "PRODUCTION",
            # "package": "great_expectations",
            "tags": [
                "arrows",
                "design",
                "flows",
            ],
            "contributors": ["@shinnyshinshin", "@abegong"],
            "requirements": [],
            "library_metadata_passed_checks": True,
            "has_full_test_suite": False,
            "manually_reviewed_code": False,
        }
    ),
    renderers=[
        ExpectationRendererDiagnostics(
            name="renderer.prescriptive",
            is_supported=True,
            is_standard=True,
            samples=[
                RendererTestDiagnostics(
                    test_title="basic_positive_test",
                    rendered_successfully=True,
                    renderered_str="a values must match this regular expression: ^a, at least 90 % of the time.",  # noqa: E501 # FIXME CoP
                )
            ],
        )
    ],
    examples=[expectation_test_data_case],
    gallery_examples=[expectation_test_data_case],
    metrics=[],
    execution_engines={
        "PandasExecutionEngine": True,
        "SqlAlchemyExecutionEngine": True,
        "SparkDFExecutionEngine": True,
    },
    tests=[],
    backend_test_result_counts=[],
    errors=[],
    coverage_score=5,
    maturity_checklist=ExpectationDiagnosticMaturityMessages(
        experimental=[
            ExpectationDiagnosticCheckMessage(
                message="AAA",
                passed=True,
            ),
            ExpectationDiagnosticCheckMessage(
                message="BBB",
                passed=False,
            ),
        ],
        beta=[],
        production=[
            ExpectationDiagnosticCheckMessage(
                message="CCC",
                passed=False,
                sub_messages=[
                    ExpectationDiagnosticCheckMessage(
                        message="ddd",
                        passed=True,
                    ),
                    ExpectationDiagnosticCheckMessage(
                        message="eee",
                        passed=False,
                    ),
                ],
            ),
        ],
    ),
)


def test__convert_checks_into_output_message():
    assert (
        edr._convert_checks_into_output_message(
            class_name="ExpectColumnValuesToEqualThree",
            maturity_level="EXPERIMENTAL",
            maturity_messages=edr.maturity_checklist,
        )
        == """\
Completeness checklist for ExpectColumnValuesToEqualThree (EXPERIMENTAL):
 ✔ AAA
   BBB
   CCC
    ✔ ddd
      eee
"""
    )


def test__count_unexpected_test_cases___with_everything_passing():
    tests = [
        ExpectationTestDiagnostics(
            test_title="positive_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="negative_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="other_negative_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
    ]
    assert edr._count_unexpected_test_cases(tests) == 0


def test__count_unexpected_test_cases__with_one_failure():
    tests = [
        ExpectationTestDiagnostics(
            test_title="positive_test_with_mostly",
            backend="pandas",
            test_passed=False,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="negative_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="other_negative_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
    ]
    assert edr._count_unexpected_test_cases(tests) == 1


def test__count_unexpected_test_cases__with_an_error():
    tests = [
        ExpectationTestDiagnostics(
            test_title="positive_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="negative_test_with_mostly",
            backend="pandas",
            test_passed=True,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="other_negative_test_with_mostly",
            backend="pandas",
            test_passed=False,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="test_that_will_error_out",
            backend="pandas",
            test_passed=False,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
        ExpectationTestDiagnostics(
            test_title="another_test_that_will_error_out",
            backend="pandas",
            test_passed=False,
            include_in_gallery=False,
            validation_result=[],
            error_diagnostics=None,
        ),
    ]
    assert edr._count_unexpected_test_cases(tests) == 3


def test__count_positive_and_negative_example_cases():
    assert edr._count_positive_and_negative_example_cases([expectation_test_data_case]) == (1, 1)


def test__check_example_cases__with_enough_test_cases_but_all_failing():
    assert ExpectationDiagnostics._check_example_cases(
        examples=[expectation_test_data_case],
        tests=[
            ExpectationTestDiagnostics(
                test_title="positive_test_exact_mostly_w_one_non_matching_value",
                backend="pandas",
                test_passed=False,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
            ExpectationTestDiagnostics(
                test_title="negative_test_insufficient_mostly_and_one_non_matching_value",
                backend="pandas",
                test_passed=False,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
        ],
    ) == ExpectationDiagnosticCheckMessage(
        **{
            "message": "Has at least one positive and negative example case, and all test cases pass",  # noqa: E501 # FIXME CoP
            "passed": False,
        }
    )


def test__check_example_cases__with_enough_test_cases_but_some_failing():
    assert ExpectationDiagnostics._check_example_cases(
        examples=[expectation_test_data_case],
        tests=[
            ExpectationTestDiagnostics(
                test_title="positive_test_exact_mostly_w_one_non_matching_value",
                backend="pandas",
                test_passed=True,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
            ExpectationTestDiagnostics(
                test_title="negative_test_insufficient_mostly_and_one_non_matching_value",
                backend="pandas",
                test_passed=False,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
        ],
    ) == ExpectationDiagnosticCheckMessage(
        **{
            "message": "Has at least one positive and negative example case, and all test cases pass",  # noqa: E501 # FIXME CoP
            "passed": False,
        }
    )


def test__check_example_cases__with_enough_test_cases_and_no_failing():
    assert ExpectationDiagnostics._check_example_cases(
        examples=[expectation_test_data_case],
        tests=[
            ExpectationTestDiagnostics(
                test_title="positive_test_exact_mostly_w_one_non_matching_value",
                backend="pandas",
                test_passed=True,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
            ExpectationTestDiagnostics(
                test_title="negative_test_insufficient_mostly_and_one_non_matching_value",
                backend="pandas",
                test_passed=True,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
        ],
    ) == ExpectationDiagnosticCheckMessage(
        **{
            "message": "Has at least one positive and negative example case, and all test cases pass",  # noqa: E501 # FIXME CoP
            "passed": True,
        }
    )


def test__check_example_cases__with_enough_not_enough_test_cases_but_no_failing():
    expectation_test_data_case = ExpectationTestDataCases(
        dataset_name="expectation_diagnostics_example_test_case",
        data=TestData(
            **{
                "a": ["aaa", "abb", "acc", "add", "bee"],
                "b": ["aaa", "abb", "acc", "bdd", None],
                "column_name with space": ["aaa", "abb", "acc", "add", "bee"],
            }
        ),
        tests=[
            ExpectationLegacyTestCaseAdapter(
                **{
                    "title": "negative_test_insufficient_mostly_and_one_non_matching_value",
                    "exact_match_out": False,
                    "in": {"column": "a", "regex": "^a", "mostly": 0.9},
                    "out": {
                        "success": False,
                        "unexpected_index_list": [4],
                        "unexpected_list": ["bee"],
                    },
                    "suppress_test_for": ["sqlite", "mssql"],
                }
            ),
        ],
    )

    assert ExpectationDiagnostics._check_example_cases(
        examples=[expectation_test_data_case],
        tests=[
            ExpectationTestDiagnostics(
                test_title="negative_test_insufficient_mostly_and_one_non_matching_value",
                backend="pandas",
                test_passed=True,
                include_in_gallery=False,
                validation_result=[],
                error_diagnostics=None,
            ),
        ],
    ) == ExpectationDiagnosticCheckMessage(
        **{
            "message": "Has at least one positive and negative example case, and all test cases pass",  # noqa: E501 # FIXME CoP
            "passed": False,
        }
    )



import shutil

import pytest

from great_expectations.data_context import get_context
from great_expectations.data_context.util import file_relative_path
from great_expectations.exceptions import InvalidConfigError


@pytest.mark.filesystem
def test_incomplete_uncommitted(tmp_path):
    """
    When a project is shared between users, it is common to have an incomplete
    uncommitted directory present. We should fail gracefully when config
    variables are missing.
    """
    local_dir = tmp_path / "root"
    fixture_path = file_relative_path(
        __file__,
        "./fixtures/contexts/incomplete_uncommitted/great_expectations",
    )
    shutil.copytree(fixture_path, local_dir)

    with pytest.raises(InvalidConfigError) as exc:
        _ = get_context(context_root_dir=local_dir)

    assert (
        "Unable to find a match for config substitution variable: "
        "`secret_validation_results_store_name`." in exc.value.message
    )
    assert (
        "See https://docs.greatexpectations.io/docs/core/configure_project_settings/configure_credentials"
        in exc.value.message
    )



from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING
from unittest.mock import patch

import pandas as pd
import pytest

import great_expectations.exceptions as gx_exceptions
import great_expectations.expectations as gxe
from great_expectations.core import ExpectationSuite
from great_expectations.core.batch import (
    BatchRequest,
)
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.result_format import ResultFormat
from great_expectations.data_context import get_context
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.datasource.fluent.pandas_datasource import PandasDatasource
from great_expectations.execution_engine import PandasExecutionEngine
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.validator.exception_info import ExceptionInfo
from great_expectations.validator.validator import Validator

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

DATASOURCE_NAME = "my_datasource"
DATA_ASSET_NAME = "IN_MEMORY_DATA_ASSET"


@pytest.fixture
def basic_datasource() -> PandasDatasource:
    context = get_context(mode="ephemeral")
    datasource = context.data_sources.add_pandas("my_datasource")
    return datasource


@pytest.fixture()
def yellow_trip_pandas_data_context(
    tmp_path_factory,
    monkeypatch,
) -> FileDataContext:
    """
    Provides a data context with a data_connector for a pandas datasource which can connect to three
    months of yellow trip taxi data in csv form. This data connector enables access to all three
    months through a BatchRequest where the "year" in batch_filter_parameters is set to "2019", or
    to individual months if the "month" in batch_filter_parameters is set to "01", "02", or "03"
    """
    project_path: str = str(tmp_path_factory.mktemp("taxi_data_context"))
    context_path: str = os.path.join(  # noqa: PTH118 # FIXME CoP
        project_path, FileDataContext.GX_DIR
    )
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(context_path, "expectations"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    data_path: str = os.path.join(context_path, "..", "data")  # noqa: PTH118 # FIXME CoP
    os.makedirs(os.path.join(data_path), exist_ok=True)  # noqa: PTH118, PTH103 # FIXME CoP
    shutil.copy(
        file_relative_path(
            __file__,
            os.path.join(  # noqa: PTH118 # FIXME CoP
                "..",
                "integration",
                "fixtures",
                "yellow_tripdata_pandas_fixture",
                FileDataContext._LEGACY_GX_DIR,
                FileDataContext.GX_YML,
            ),
        ),
        str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    shutil.copy(
        file_relative_path(
            __file__,
            os.path.join(  # noqa: PTH118 # FIXME CoP
                "..",
                "test_sets",
                "taxi_yellow_tripdata_samples",
                "yellow_tripdata_sample_2019-01.csv",
            ),
        ),
        str(
            os.path.join(  # noqa: PTH118 # FIXME CoP
                context_path, "..", "data", "yellow_tripdata_sample_2019-01.csv"
            )
        ),
    )
    shutil.copy(
        file_relative_path(
            __file__,
            os.path.join(  # noqa: PTH118 # FIXME CoP
                "..",
                "test_sets",
                "taxi_yellow_tripdata_samples",
                "yellow_tripdata_sample_2019-02.csv",
            ),
        ),
        str(
            os.path.join(  # noqa: PTH118 # FIXME CoP
                context_path, "..", "data", "yellow_tripdata_sample_2019-02.csv"
            )
        ),
    )
    shutil.copy(
        file_relative_path(
            __file__,
            os.path.join(  # noqa: PTH118 # FIXME CoP
                "..",
                "test_sets",
                "taxi_yellow_tripdata_samples",
                "yellow_tripdata_sample_2019-03.csv",
            ),
        ),
        str(
            os.path.join(  # noqa: PTH118 # FIXME CoP
                context_path, "..", "data", "yellow_tripdata_sample_2019-03.csv"
            )
        ),
    )

    context = get_context(context_root_dir=context_path)
    assert context.root_directory == context_path

    return context


@pytest.mark.big
def test_validator_default_expectation_args__pandas(basic_datasource: PandasDatasource):
    asset = basic_datasource.add_dataframe_asset(
        "my_asset",
    )
    batch_definition = asset.add_batch_definition_whole_dataframe(
        name="my batch definition",
    )
    batch = batch_definition.get_batch(
        batch_parameters={
            "dataframe": pd.DataFrame({"a": [1, 5, 22, 3, 5, 10], "b": [1, 2, 3, 4, 5, None]})
        }
    )

    my_validator = Validator(execution_engine=PandasExecutionEngine(), batches=[batch])

    assert my_validator.get_default_expectation_arguments() == {
        "catch_exceptions": False,
        "result_format": ResultFormat.BASIC,
    }


@pytest.fixture()
def multi_batch_taxi_validator_ge_cloud_mode(
    yellow_trip_pandas_data_context,
) -> Validator:
    context = yellow_trip_pandas_data_context
    context._cloud_mode = True

    suite = ExpectationSuite(
        name="validating_taxi_data",
        expectations=[
            ExpectationConfiguration(
                type="expect_column_values_to_be_between",
                kwargs={
                    "column": "passenger_count",
                    "min_value": 0,
                    "max_value": 99,
                    "result_format": "BASIC",
                },
                meta={"notes": "This is an expectation."},
                id="0faf94a9-f53a-41fb-8e94-32f218d4a774",
            )
        ],
        meta={"notes": "This is an expectation suite."},
    )

    multi_batch_request: BatchRequest = BatchRequest(
        datasource_name="taxi_pandas",
        data_connector_name="monthly",
        data_asset_name="my_reports",
        data_connector_query={"batch_filter_parameters": {"year": "2019"}},
    )

    validator_multi_batch: Validator = context.get_validator(
        batch_request=multi_batch_request, expectation_suite=suite
    )

    return validator_multi_batch


@pytest.mark.big
def test_graph_validate(in_memory_runtime_context, basic_datasource: PandasDatasource):
    asset = basic_datasource.add_dataframe_asset("my_asset")
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(
        batch_parameters={
            "dataframe": pd.DataFrame({"a": [1, 5, 22, 3, 5, 10], "b": [1, 2, 3, 4, 5, None]})
        },
    )

    expectation_configuration = ExpectationConfiguration(
        type="expect_column_value_z_scores_to_be_less_than",
        kwargs={
            "column": "b",
            "mostly": 0.9,
            "threshold": 4.0,
            "double_sided": True,
        },
    )
    result = Validator(
        execution_engine=PandasExecutionEngine(),
        data_context=in_memory_runtime_context,
        batches=[batch],
    ).graph_validate(configurations=[expectation_configuration])
    assert result == [
        ExpectationValidationResult(
            success=True,
            expectation_config=None,
            meta={},
            result={
                "element_count": 6,
                "unexpected_count": 0,
                "unexpected_percent": 0.0,
                "partial_unexpected_list": [],
                "missing_count": 1,
                "missing_percent": 16.666666666666664,
                "unexpected_percent_nonmissing": 0.0,
            },
            exception_info=None,
        )
    ]


# Tests that runtime configuration actually works during graph validation
@pytest.mark.big
def test_graph_validate_with_runtime_config(
    in_memory_runtime_context, basic_datasource: PandasDatasource
):
    asset = basic_datasource.add_dataframe_asset(
        "my_asset",
    )
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(
        {
            "dataframe": pd.DataFrame(
                {"a": [1, 5, 22, 3, 5, 10, 2, 3], "b": [97, 332, 3, 4, 5, 6, 7, None]}
            )
        },
    )

    expectation_configuration = ExpectationConfiguration(
        type="expect_column_value_z_scores_to_be_less_than",
        kwargs={"column": "b", "mostly": 1.0, "threshold": 2.0, "double_sided": True},
    )
    result = Validator(
        execution_engine=PandasExecutionEngine(),
        data_context=in_memory_runtime_context,
        batches=[batch],
    ).graph_validate(
        configurations=[expectation_configuration],
        runtime_configuration={"result_format": "COMPLETE"},
    )

    assert result == [
        ExpectationValidationResult(
            success=False,
            meta={},
            result={
                "element_count": 8,
                "unexpected_count": 1,
                "unexpected_percent": 14.285714285714285,
                "partial_unexpected_list": [332.0],
                "missing_count": 1,
                "missing_percent": 12.5,
                "unexpected_percent_total": 12.5,
                "unexpected_percent_nonmissing": 14.285714285714285,
                "partial_unexpected_index_list": [1],
                "partial_unexpected_counts": [{"value": 332.0, "count": 1}],
                "unexpected_list": [332.0],
                "unexpected_index_list": [1],
            },
            expectation_config=gxe.ExpectColumnValueZScoresToBeLessThan(
                column="b",
                mostly=1.0,
                threshold=2.0,
                double_sided=True,
            ).configuration,
            exception_info=None,
        )
    ]


@pytest.mark.big
def test_graph_validate_with_exception(basic_datasource: PandasDatasource, mocker: MockerFixture):
    # noinspection PyUnusedLocal
    def mock_error(*args, **kwargs):
        raise Exception("Mock Error")  # noqa: TRY002 # FIXME CoP

    df = pd.DataFrame({"a": [1, 5, 22, 3, 5, 10], "b": [1, 2, 3, 4, 5, None]})
    asset = basic_datasource.add_dataframe_asset("my_asset")
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(batch_parameters={"dataframe": df})

    expectation_configuration = ExpectationConfiguration(
        type="expect_column_value_z_scores_to_be_less_than",
        kwargs={
            "column": "b",
            "mostly": 0.9,
            "threshold": 4,
            "double_sided": True,
        },
    )

    execution_engine = PandasExecutionEngine()
    validator = Validator(execution_engine=execution_engine, batches=[batch])

    # TODO: Convert this to actually mock an exception being thrown
    # graph = ValidationGraph(execution_engine=execution_engine)
    # graph.build_metric_dependency_graph = mock_error  # type: ignore[method-assign] # FIXME CoP

    result = validator.graph_validate(configurations=[expectation_configuration])

    assert len(result) == 1
    assert result[0].expectation_config is not None


@pytest.mark.big
def test_graph_validate_with_bad_config_catch_exceptions_false(
    in_memory_runtime_context, basic_datasource: PandasDatasource
):
    df = pd.DataFrame({"a": [1, 5, 22, 3, 5, 10], "b": [1, 2, 3, 4, 5, None]})
    asset = basic_datasource.add_dataframe_asset("my_asset")
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(batch_parameters={"dataframe": df})

    expectation_configuration = ExpectationConfiguration(
        type="expect_column_max_to_be_between",
        kwargs={"column": "not_in_table", "min_value": 1, "max_value": 29},
    )
    with pytest.raises(gx_exceptions.MetricResolutionError) as eee:
        # noinspection PyUnusedLocal
        _ = Validator(
            execution_engine=PandasExecutionEngine(),
            data_context=in_memory_runtime_context,
            batches=[batch],
        ).graph_validate(
            configurations=[expectation_configuration],
            runtime_configuration={
                "catch_exceptions": False,
                "result_format": {"result_format": "BASIC"},
            },
        )
    assert str(eee.value) == 'Error: The column "not_in_table" in BatchData does not exist.'


@pytest.mark.parametrize(
    "value_set, expected",
    [
        (list(range(2)), False),  # value set won't pass for the actual data
        (list(range(5)), True),  # value set will pass for the actual data
    ],
)
@pytest.mark.big
def test_validator_validate_substitutes_suite_parameters(
    value_set: list[int],
    expected: bool,
):
    """Integration test to ensure suite parameters are respected when validating.
    The setup here is to provide very simple data, and a variety of evaluation_parameter inputs,
    just checking for result.success as a proxy for the evaluation_parameter being respected.
    """

    # Arrange
    context = get_context(mode="ephemeral")
    suite_name = "my_suite"
    column_name = "my_column"
    datasource = context.data_sources.add_pandas(name="my_datasource")
    asset = datasource.add_dataframe_asset(
        "my_asset",
    )
    suite = context.suites.add(ExpectationSuite(suite_name))
    suite.add_expectation(
        gxe.ExpectColumnDistinctValuesToBeInSet(
            column=column_name, value_set={"$PARAMETER": "value_set"}
        )
    )
    validator = context.get_validator(
        batch_request=asset.build_batch_request(
            options={"dataframe": pd.DataFrame({column_name: [0, 1, 2, 3, 4]})}
        ),
        expectation_suite_name=suite_name,
    )

    # Act
    result = validator.validate(suite_parameters={"value_set": value_set})
    assert isinstance(result, ExpectationSuiteValidationResult)
    suite_parameters_used = result.results[0]["expectation_config"]["kwargs"]["value_set"]

    # Assert
    assert suite_parameters_used == value_set
    assert result.success == expected


@pytest.mark.unit
@pytest.mark.parametrize("result_format", ["BOOLEAN_ONLY", {"result_format": "BOOLEAN_ONLY"}])
def test_rendered_content_bool_only_respected(result_format: str | dict):
    context = get_context(mode="ephemeral")
    csv_asset = context.data_sources.pandas_default.add_dataframe_asset(
        "df",
    )
    batch_request = csv_asset.build_batch_request(
        options={
            "dataframe": pd.DataFrame(
                data=[1, 2, 3],
                columns=["numbers_i_can_count_to"],
            )
        }
    )
    expectation_suite_name = "test_result_format_suite"
    context.suites.add(ExpectationSuite(name=expectation_suite_name))

    validator = context.get_validator(
        batch_request=batch_request,
        expectation_suite_name=expectation_suite_name,
    )

    expectation_validation_result = validator.expect_column_max_to_be_between(
        column="numbers_i_can_count_to",
        min_value=1000,
        max_value=10000,
        result_format=result_format,
    )
    assert expectation_validation_result.result == {}


@pytest.mark.unit
def test___dir___contains_expectation_impls(
    validator_with_mock_execution_engine: Validator,
) -> None:
    validator = validator_with_mock_execution_engine

    attrs = validator.__dir__()
    expectation_impls = list(filter(lambda a: a.startswith("expect_"), attrs))
    assert len(expectation_impls) > 0


@pytest.mark.unit
def test_show_progress_bars_property_and_setter(
    validator_with_mock_execution_engine: Validator,
) -> None:
    validator = validator_with_mock_execution_engine

    assert validator.metrics_calculator.show_progress_bars is True
    validator.metrics_calculator.show_progress_bars = False
    assert validator.metrics_calculator.show_progress_bars is False


@pytest.mark.unit
def test_expose_dataframe_methods_property_and_setter(
    validator_with_mock_execution_engine: Validator,
) -> None:
    validator = validator_with_mock_execution_engine

    assert validator.expose_dataframe_methods is False
    validator.expose_dataframe_methods = True
    assert validator.expose_dataframe_methods is True


@pytest.mark.unit
def test___get_attr___retrieves_existing_expectation(
    validator_with_mock_execution_engine: Validator,
) -> None:
    validator = validator_with_mock_execution_engine

    # Does not raise error if properly registered
    # Avoiding invocation to only test registration (and not actual expectation)
    _ = validator.expect_column_max_to_be_between


@pytest.mark.unit
def test__get_attr___raises_attribute_error_with_invalid_attr(
    validator_with_mock_execution_engine: Validator,
) -> None:
    validator = validator_with_mock_execution_engine

    with pytest.raises(AttributeError) as e:
        _ = validator.my_fake_attr

    assert "'Validator'  object has no attribute 'my_fake_attr'" in str(e.value)


@pytest.mark.big
def test_graph_validate_with_two_expectations_and_first_expectation_without_additional_configuration(  # noqa: E501 # FIXME CoP
    in_memory_runtime_context, basic_datasource: PandasDatasource
):
    in_memory_runtime_context.data_sources.all()["my_datasource"] = basic_datasource
    df = pd.DataFrame(
        [
            "A",
            "B",
            "B",
            "C",
            "C",
            "C",
            "D",
            "D",
            "D",
            "D",
            "E",
            "E",
            "E",
            "E",
            "E",
            "F",
            "F",
            "F",
            "F",
            "F",
            "F",
            "G",
            "G",
            "G",
            "G",
            "G",
            "G",
            "G",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
        ],
        columns=["var"],
    )
    asset = basic_datasource.add_dataframe_asset("my_asset")
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(batch_parameters={"dataframe": df})

    expectation_configuration_expect_column_values_to_be_null = ExpectationConfiguration(
        type="expect_column_values_to_be_null",
        kwargs={
            "column": "var",
        },
    )

    expectation_configuration_expect_column_values_to_be_in_set = ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "var",
            "value_set": ["B", "C", "D", "F", "G", "H"],
            "result_format": {"result_format": "COMPLETE"},
        },
    )
    result = Validator(
        execution_engine=PandasExecutionEngine(),
        data_context=in_memory_runtime_context,
        batches=[batch],
    ).graph_validate(
        configurations=[
            expectation_configuration_expect_column_values_to_be_null,
            expectation_configuration_expect_column_values_to_be_in_set,
        ]
    )

    assert result == [
        ExpectationValidationResult(
            success=False,
            expectation_config=ExpectationConfiguration(
                "expect_column_values_to_be_null",
                kwargs={
                    "column": "var",
                },
            ),
            meta={},
            result={
                "element_count": 36,
                "unexpected_count": 36,
                "unexpected_percent": 100.0,
                "partial_unexpected_list": [
                    "A",
                    "B",
                    "B",
                    "C",
                    "C",
                    "C",
                    "D",
                    "D",
                    "D",
                    "D",
                    "E",
                    "E",
                    "E",
                    "E",
                    "E",
                    "F",
                    "F",
                    "F",
                    "F",
                    "F",
                ],
            },
            exception_info=None,
        ),
        ExpectationValidationResult(
            success=False,
            expectation_config=ExpectationConfiguration(
                "expect_column_values_to_be_in_set",
                kwargs={
                    "column": "var",
                    "value_set": ["B", "C", "D", "F", "G", "H"],
                    "result_format": {"result_format": "COMPLETE"},
                },
            ),
            meta={},
            result={
                "element_count": 36,
                "unexpected_count": 6,
                "unexpected_percent": 16.666666666666664,
                "partial_unexpected_list": ["A", "E", "E", "E", "E", "E"],
                "missing_count": 0,
                "missing_percent": 0.0,
                "unexpected_percent_total": 16.666666666666664,
                "unexpected_percent_nonmissing": 16.666666666666664,
                "partial_unexpected_index_list": [0, 10, 11, 12, 13, 14],
                "partial_unexpected_counts": [
                    {"value": "E", "count": 5},
                    {"value": "A", "count": 1},
                ],
                "unexpected_list": ["A", "E", "E", "E", "E", "E"],
                "unexpected_index_list": [0, 10, 11, 12, 13, 14],
            },
            exception_info=None,
        ),
    ]


@pytest.mark.big
def test_graph_validate_with_two_expectations_and_first_expectation_with_result_format_complete(
    in_memory_runtime_context, basic_datasource: PandasDatasource
):
    in_memory_runtime_context.data_sources.all()["my_datasource"] = basic_datasource
    df = pd.DataFrame(
        [
            "A",
            "B",
            "B",
            "C",
            "C",
            "C",
            "D",
            "D",
            "D",
            "D",
            "E",
            "E",
            "E",
            "E",
            "E",
            "F",
            "F",
            "F",
            "F",
            "F",
            "F",
            "G",
            "G",
            "G",
            "G",
            "G",
            "G",
            "G",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
            "H",
        ],
        columns=["var"],
    )

    asset = basic_datasource.add_dataframe_asset("my_asset")
    batch_definition = asset.add_batch_definition_whole_dataframe("my batch definition")
    batch = batch_definition.get_batch(batch_parameters={"dataframe": df})

    expectation_configuration_expect_column_values_to_be_null = ExpectationConfiguration(
        type="expect_column_values_to_be_null",
        kwargs={
            "column": "var",
        },
    )

    expectation_configuration_expect_column_values_to_be_in_set = ExpectationConfiguration(
        type="expect_column_values_to_be_in_set",
        kwargs={
            "column": "var",
            "value_set": ["B", "C", "D", "F", "G", "H"],
            "result_format": {"result_format": "COMPLETE"},
        },
    )
    result = Validator(
        execution_engine=PandasExecutionEngine(),
        data_context=in_memory_runtime_context,
        batches=[batch],
    ).graph_validate(
        configurations=[
            expectation_configuration_expect_column_values_to_be_in_set,
            expectation_configuration_expect_column_values_to_be_null,
        ]
    )

    assert result == [
        ExpectationValidationResult(
            success=False,
            expectation_config=ExpectationConfiguration(
                "expect_column_values_to_be_in_set",
                kwargs={
                    "column": "var",
                    "value_set": ["B", "C", "D", "F", "G", "H"],
                    "result_format": {"result_format": "COMPLETE"},
                },
            ),
            meta={},
            result={
                "element_count": 36,
                "unexpected_count": 6,
                "unexpected_percent": 16.666666666666664,
                "partial_unexpected_list": ["A", "E", "E", "E", "E", "E"],
                "missing_count": 0,
                "missing_percent": 0.0,
                "unexpected_percent_total": 16.666666666666664,
                "unexpected_percent_nonmissing": 16.666666666666664,
                "partial_unexpected_index_list": [0, 10, 11, 12, 13, 14],
                "partial_unexpected_counts": [
                    {"value": "E", "count": 5},
                    {"value": "A", "count": 1},
                ],
                "unexpected_list": ["A", "E", "E", "E", "E", "E"],
                "unexpected_index_list": [0, 10, 11, 12, 13, 14],
            },
            exception_info=None,
        ),
        ExpectationValidationResult(
            success=False,
            expectation_config=ExpectationConfiguration(
                "expect_column_values_to_be_null",
                kwargs={
                    "column": "var",
                },
            ),
            meta={},
            result={
                "element_count": 36,
                "unexpected_count": 36,
                "unexpected_percent": 100.0,
                "partial_unexpected_list": [
                    "A",
                    "B",
                    "B",
                    "C",
                    "C",
                    "C",
                    "D",
                    "D",
                    "D",
                    "D",
                    "E",
                    "E",
                    "E",
                    "E",
                    "E",
                    "F",
                    "F",
                    "F",
                    "F",
                    "F",
                ],
            },
            exception_info=None,
        ),
    ]


@pytest.mark.unit
def test_validator_with_exception_info_in_result():
    get_context()
    validator = Validator(execution_engine=PandasExecutionEngine())

    mock_resolved_metrics = []

    metric_id = (
        "table.column_types",
        "a351bbf72b281f0b7a62dbbf3599ce5c",
        "include_nested=True",
    )
    exception_message = "Danger Will Robinson! Danger!"
    exception_traceback = 'Traceback (most recent call last):\n File "lostinspace.py", line 42, in <module>\n    raise Exception("Danger Will Robinson! Danger!")\nException: Danger Will Robinson! Danger!'  # noqa: E501 # FIXME CoP

    mock_aborted_metrics_info = {
        metric_id: {
            "exception_info": ExceptionInfo(
                exception_traceback=exception_traceback,
                exception_message=exception_message,
            ),
        }
    }

    with patch.object(
        validator._metrics_calculator,
        "resolve_validation_graph",
        return_value=(mock_resolved_metrics, mock_aborted_metrics_info),
    ):
        result = validator.graph_validate(
            configurations=[
                ExpectationConfiguration(
                    type="expect_column_values_to_be_unique",
                    kwargs={
                        "column": "animals",
                    },
                ),
                ExpectationConfiguration(
                    type="expect_column_values_to_be_in_set",
                    kwargs={
                        "column": "animals",
                        "value_set": ["cat", "fish", "dog"],
                    },
                ),
                ExpectationConfiguration(
                    type="expect_column_values_to_not_be_null",
                    kwargs={
                        "column": "animals",
                    },
                ),
            ],
            runtime_configuration={"result_format": "COMPLETE"},
        )
        assert len(result) == 3  # One to one mapping of Expectations to results
        for evr in result:
            assert evr.exception_info is not None
            assert evr.exception_info[str(metric_id)].exception_traceback == exception_traceback
            assert evr.exception_info[str(metric_id)].exception_message == exception_message



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from great_expectations.core.data_context_key import StringKey
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store import ValidationDefinitionStore
from great_expectations.data_context.types.resource_identifiers import GXCloudIdentifier

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.ephemeral_data_context import (
        EphemeralDataContext,
    )
    from tests.datasource.fluent._fake_cloud_api import CloudDetails


@pytest.fixture
def ephemeral_store():
    return ValidationDefinitionStore(store_name="ephemeral_validation_definition_store")


@pytest.fixture
def file_backed_store(tmp_path):
    base_directory = tmp_path / "base_dir"
    return ValidationDefinitionStore(
        store_name="file_backed_validation_definition_store",
        store_backend={
            "class_name": "TupleFilesystemStoreBackend",
            "base_directory": base_directory,
        },
    )


@pytest.fixture
def cloud_backed_store(cloud_details: CloudDetails):
    return ValidationDefinitionStore(
        store_name="cloud_backed_validation_definition_store",
        store_backend={
            "class_name": "GXCloudStoreBackend",
            "ge_cloud_resource_type": GXCloudRESTResource.VALIDATION_DEFINITION,
            "ge_cloud_credentials": {
                "access_token": cloud_details.access_token,
                "organization_id": cloud_details.org_id,
            },
        },
    )


@pytest.fixture
def validation_definition(
    in_memory_runtime_context: EphemeralDataContext,
) -> ValidationDefinition:
    context = in_memory_runtime_context
    batch_definition = (
        context.data_sources.add_pandas("my_datasource")
        .add_csv_asset("my_asset", "data.csv")  # type: ignore[arg-type] # FIXME CoP
        .add_batch_definition("my_batch_definition")
    )
    return ValidationDefinition(
        name="my_validation",
        data=batch_definition,
        suite=context.suites.add(ExpectationSuite(name="my_suite")),
    )


@pytest.mark.parametrize("store_fixture", ["ephemeral_store", "file_backed_store"])
@pytest.mark.unit
def test_add(request, store_fixture: str, validation_definition: ValidationDefinition):
    store: ValidationDefinitionStore = request.getfixturevalue(store_fixture)

    key = StringKey(key="my_validation")

    assert not validation_definition.id
    store.add(key=key, value=validation_definition)
    assert validation_definition.id

    assert store.get(key) == validation_definition


@pytest.mark.cloud
def test_add_cloud(
    cloud_backed_store: ValidationDefinitionStore, validation_definition: ValidationDefinition
):
    store = cloud_backed_store

    id = "5a8ada9f-5b71-461b-b1af-f1d93602a156"
    name = "my_validation"
    key = GXCloudIdentifier(
        resource_type=GXCloudRESTResource.VALIDATION_DEFINITION,
        id=id,
        resource_name=name,
    )

    with mock.patch("requests.Session.put", autospec=True) as mock_put:
        store.add(key=key, value=validation_definition)

    mock_put.assert_called_once_with(
        mock.ANY,  # requests Session
        f"https://api.greatexpectations.io/api/v1/organizations/12345678-1234-5678-1234-567812345678/validation-definitions/{id}",
        json={
            "data": {
                "name": name,
                "data": {
                    # NOTE: IDs here are unrealistic because the validation_definition
                    # was created via the in-memory store
                    "batch_definition": {"id": mock.ANY, "name": "my_batch_definition"},
                    "asset": {"id": mock.ANY, "name": "my_asset"},
                    "datasource": {"id": mock.ANY, "name": "my_datasource"},
                },
                "suite": {
                    "name": "my_suite",
                    "id": mock.ANY,
                },
            },
        },
    )


@pytest.mark.parametrize("store_fixture", ["ephemeral_store", "file_backed_store"])
@pytest.mark.unit
def test_get_key(request, store_fixture: str):
    store: ValidationDefinitionStore = request.getfixturevalue(store_fixture)

    name = "my_validation"
    assert store.get_key(name=name) == StringKey(key=name)


@pytest.mark.cloud
def test_get_key_cloud(cloud_backed_store: ValidationDefinitionStore):
    key = cloud_backed_store.get_key(name="my_validation")
    assert key.resource_type == GXCloudRESTResource.VALIDATION_DEFINITION  # type: ignore[union-attr] # FIXME CoP
    assert key.resource_name == "my_validation"  # type: ignore[union-attr] # FIXME CoP


_VALIDATION_ID = "a4sdfd-64c8-46cb-8f7e-03c12cea1d67"
_VALIDATION_DEFINITION = {
    "name": "my_validation",
    "data": {
        "datasource": {
            "name": "my_datasource",
            "id": "a758816-64c8-46cb-8f7e-03c12cea1d67",
        },
        "asset": {
            "name": "my_asset",
            "id": "b5s8816-64c8-46cb-8f7e-03c12cea1d67",
        },
        "batch_definition": {
            "name": "my_batch_definition",
            "id": "3a758816-64c8-46cb-8f7e-03c12cea1d67",
        },
    },
    "suite": {
        "name": "my_suite",
        "id": "8r2g816-64c8-46cb-8f7e-03c12cea1d67",
    },
}


@pytest.mark.cloud
@pytest.mark.parametrize(
    "response_json",
    [
        pytest.param(
            {
                "data": {
                    **_VALIDATION_DEFINITION,
                    "id": _VALIDATION_ID,
                }
            },
            id="single_validation_definition",
        ),
        pytest.param(
            {
                "data": [
                    {
                        **_VALIDATION_DEFINITION,
                        "id": _VALIDATION_ID,
                    }
                ]
            },
            id="list_with_single_validation_definition",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict_success(response_json: dict):
    actual = ValidationDefinitionStore.gx_cloud_response_json_to_object_dict(response_json)
    expected = {**_VALIDATION_DEFINITION, "id": _VALIDATION_ID}
    assert actual == expected


@pytest.mark.cloud
@pytest.mark.parametrize(
    "response_json, error_substring",
    [
        pytest.param(
            {
                "data": [],
            },
            "Cannot parse empty data",
            id="empty_list",
        ),
        pytest.param(
            {
                "data": [
                    {
                        "id": _VALIDATION_ID,
                        "attributes": {
                            "validation_definition": _VALIDATION_DEFINITION,
                        },
                    },
                    {
                        "id": _VALIDATION_ID,
                        "attributes": {
                            "validation_definition": _VALIDATION_DEFINITION,
                        },
                    },
                ],
            },
            "Cannot parse multiple items",
            id="list_with_multiple_validation_definitions",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict_failure(response_json: dict, error_substring: str):
    with pytest.raises(ValueError, match=f"{error_substring}*."):
        ValidationDefinitionStore.gx_cloud_response_json_to_object_dict(response_json)



from __future__ import annotations

from typing import Type

from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.dynamic_pandas import _generate_pandas_data_asset_models

_PANDAS_FILE_TYPE_READER_METHOD_UNSUPPORTED_LIST = (
    # "read_csv",
    # "read_json",
    # "read_excel",
    # "read_parquet",
    "read_clipboard",  # not path based
    # "read_feather",
    # "read_fwf",
    "read_gbq",  # not path based
    # "read_hdf",
    # "read_html",
    # "read_orc",
    # "read_pickle",
    # "read_sas",  # invalid json schema
    # "read_spss",
    "read_sql",  # not path based & type-name conflict
    "read_sql_query",  # not path based
    "read_sql_table",  # not path based
    "read_table",  # type-name conflict
    # "read_xml",
)
_FILE_PATH_ASSET_MODELS = _generate_pandas_data_asset_models(
    FileDataAsset,
    blacklist=_PANDAS_FILE_TYPE_READER_METHOD_UNSUPPORTED_LIST,
    use_docstring_from_method=True,
    skip_first_param=True,
)
CSVAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("csv", FileDataAsset)
ExcelAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("excel", FileDataAsset)
FWFAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("fwf", FileDataAsset)
JSONAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("json", FileDataAsset)
ORCAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("orc", FileDataAsset)
ParquetAsset: Type[FileDataAsset] = _FILE_PATH_ASSET_MODELS.get("parquet", FileDataAsset)



import pathlib
import uuid
from logging import Logger
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
    Generator,
    List,
    NamedTuple,
    Optional,
    Type,
    Union,
    overload,
)

from typing_extensions import TypeAlias, override

from great_expectations.compatibility import pydantic
from great_expectations.data_context import (
    AbstractDataContext as GXDataContext,
)
from great_expectations.datasource.datasource_dict import DatasourceDict
from great_expectations.datasource.fluent import (
    DatabricksSQLDatasource,
    FabricPowerBIDatasource,
    PandasAzureBlobStorageDatasource,
    PandasDatasource,
    PandasDBFSDatasource,
    PandasFilesystemDatasource,
    PandasGoogleCloudStorageDatasource,
    PandasS3Datasource,
    PostgresDatasource,
    SnowflakeDatasource,
    SparkAzureBlobStorageDatasource,
    SparkDatasource,
    SparkDBFSDatasource,
    SparkFilesystemDatasource,
    SparkGoogleCloudStorageDatasource,
    SparkS3Datasource,
    SQLDatasource,
    SqliteDatasource,
)
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.databricks_sql_datasource import DatabricksDsn
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    Datasource,
)
from great_expectations.datasource.fluent.redshift_datasource import RedshiftDsn
from great_expectations.datasource.fluent.snowflake_datasource import (
    ConnectionDetails as SnowflakeConnectionDetails,
)
from great_expectations.datasource.fluent.snowflake_datasource import (
    SnowflakeDsn,
)
from great_expectations.datasource.fluent.spark_datasource import SparkConfig
from great_expectations.datasource.fluent.sqlite_datasource import SqliteDsn
from great_expectations.datasource.fluent.type_lookup import TypeLookup

SourceFactoryFn: TypeAlias = Callable[..., Datasource]
logger: Logger
DEFAULT_PANDAS_DATASOURCE_NAME: Final[str]
DEFAULT_PANDAS_DATA_ASSET_NAME: Final[str]

class DefaultPandasDatasourceError(Exception): ...
class TypeRegistrationError(TypeError): ...

class _FieldDetails(NamedTuple):
    default_value: Any
    type_annotation: Type

def _get_field_details(model: Type[pydantic.BaseModel], field_name: str) -> _FieldDetails: ...

class DataSourceManager:
    type_lookup: ClassVar[TypeLookup]
    def __init__(self, data_context: GXDataContext) -> None: ...
    @classmethod
    def register_datasource(
        cls,
        ds_type: Type[Datasource],
    ) -> None: ...
    @property
    def pandas_default(self) -> PandasDatasource: ...
    @property
    def factories(self) -> List[str]: ...
    def __getattr__(self, attr_name: str): ...
    @override
    def __dir__(self) -> List[str]: ...
    def add_pandas(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
    ) -> PandasDatasource: ...
    def update_pandas(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
    ) -> PandasDatasource: ...
    def add_or_update_pandas(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
    ) -> PandasDatasource: ...
    def delete_pandas(
        self,
        name: str,
    ) -> None: ...
    def add_pandas_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasFilesystemDatasource: ...
    def update_pandas_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasFilesystemDatasource: ...
    def add_or_update_pandas_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasFilesystemDatasource: ...
    def delete_pandas_filesystem(
        self,
        name: str,
    ) -> None: ...
    def add_pandas_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasDBFSDatasource: ...
    def update_pandas_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasDBFSDatasource: ...
    def add_or_update_pandas_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> PandasDBFSDatasource: ...
    def delete_pandas_dbfs(
        self,
        name: str,
    ) -> None: ...
    def add_pandas_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasS3Datasource: ...
    def update_pandas_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasS3Datasource: ...
    def add_or_update_pandas_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasS3Datasource: ...
    def delete_pandas_s3(
        self,
        name: str,
    ) -> None: ...
    def add_pandas_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasGoogleCloudStorageDatasource: ...
    def update_pandas_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasGoogleCloudStorageDatasource: ...
    def add_or_update_pandas_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> PandasGoogleCloudStorageDatasource: ...
    def delete_pandas_gcs(
        self,
        name: str,
    ) -> None: ...
    def add_pandas_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        azure_options: dict[str, Any] = ...,
    ) -> PandasAzureBlobStorageDatasource: ...
    def update_pandas_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        azure_options: dict[str, Any] = ...,
    ) -> PandasAzureBlobStorageDatasource: ...
    def add_or_update_pandas_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        azure_options: dict[str, Any] = ...,
    ) -> PandasAzureBlobStorageDatasource: ...
    def delete_pandas_abs(
        self,
        name: str,
    ) -> None: ...
    def add_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, str] = ...,
        create_temp_table: bool = True,
    ) -> SQLDatasource: ...
    def update_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, str] = ...,
        create_temp_table: bool = True,
    ) -> SQLDatasource: ...
    def add_or_update_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, str] = ...,
        create_temp_table: bool = True,
    ) -> SQLDatasource: ...
    def delete_sql(
        self,
        name: str,
    ) -> None: ...
    def add_postgres(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, pydantic.networks.PostgresDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def update_postgres(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, pydantic.networks.PostgresDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def add_or_update_postgres(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, pydantic.networks.PostgresDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def delete_postgres(
        self,
        name: str,
    ) -> None: ...
    def add_redshift(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, RedshiftDsn] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def update_redshift(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, RedshiftDsn] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def add_or_update_redshift(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, RedshiftDsn] = ...,
        create_temp_table: bool = True,
    ) -> PostgresDatasource: ...
    def delete_redshift(
        self,
        name: str,
    ) -> None: ...
    def add_spark(
        self,
        name: str,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
    ) -> SparkDatasource: ...
    def update_spark(
        self,
        name: str,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
    ) -> SparkDatasource: ...
    def add_or_update_spark(
        self,
        name: str,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
    ) -> SparkDatasource: ...
    def delete_spark(
        self,
        name: str,
    ) -> None: ...
    def add_spark_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkFilesystemDatasource: ...
    def update_spark_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkFilesystemDatasource: ...
    def add_or_update_spark_filesystem(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkFilesystemDatasource: ...
    def delete_spark_filesystem(
        self,
        name: str,
    ) -> None: ...
    def add_spark_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkDBFSDatasource: ...
    def update_spark_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkDBFSDatasource: ...
    def add_or_update_spark_dbfs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        base_directory: pathlib.Path = ...,
        data_context_root_directory: Union[pathlib.Path, None] = ...,
    ) -> SparkDBFSDatasource: ...
    def delete_spark_dbfs(
        self,
        name: str,
    ) -> None: ...
    def add_spark_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkS3Datasource: ...
    def update_spark_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkS3Datasource: ...
    def add_or_update_spark_s3(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket: str = ...,
        boto3_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkS3Datasource: ...
    def delete_spark_s3(
        self,
        name: str,
    ) -> None: ...
    def add_spark_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkGoogleCloudStorageDatasource: ...
    def update_spark_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkGoogleCloudStorageDatasource: ...
    def add_or_update_spark_gcs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        bucket_or_name: str = ...,
        gcs_options: dict[str, Union[ConfigStr, Any]] = ...,
    ) -> SparkGoogleCloudStorageDatasource: ...
    def delete_spark_gcs(
        self,
        name: str,
    ) -> None: ...
    def add_spark_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        azure_options: dict[str, Any] = ...,
    ) -> SparkAzureBlobStorageDatasource: ...
    def update_spark_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        azure_options: dict[str, Any] = ...,
    ) -> SparkAzureBlobStorageDatasource: ...
    def add_or_update_spark_abs(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        spark_config: SparkConfig | None = None,
        force_reuse_spark_context: bool = True,
        persist: bool = True,
        azure_options: dict[str, Any] = ...,
    ) -> SparkAzureBlobStorageDatasource: ...
    def delete_spark_abs(
        self,
        name: str,
    ) -> None: ...
    def add_sqlite(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, SqliteDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> SqliteDatasource: ...
    def update_sqlite(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, SqliteDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> SqliteDatasource: ...
    def add_or_update_sqlite(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, SqliteDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> SqliteDatasource: ...
    def delete_sqlite(
        self,
        name: str,
    ) -> None: ...
    @overload
    def add_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: Union[
            ConfigStr, SnowflakeDsn, str, SnowflakeConnectionDetails, dict[str, str]
        ] = ...,
        create_temp_table: bool = ...,
        account: None = ...,
        user: None = ...,
        password: None = ...,
        database: None = ...,
        schema: None = ...,
        warehouse: None = ...,
        role: None = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    @overload
    def add_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: None = ...,
        create_temp_table: bool = ...,
        account: str = ...,
        user: str = ...,
        password: Union[ConfigStr, str] = ...,
        database: str = ...,
        schema: str = ...,
        warehouse: Optional[str] = ...,
        role: Optional[str] = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    @overload
    def update_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: Union[
            ConfigStr, SnowflakeDsn, str, SnowflakeConnectionDetails, dict[str, str]
        ] = ...,
        create_temp_table: bool = ...,
        account: None = ...,
        user: None = ...,
        password: None = ...,
        database: None = ...,
        schema: None = ...,
        warehouse: None = ...,
        role: None = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    @overload
    def update_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: None = ...,
        create_temp_table: bool = ...,
        account: str = ...,
        user: str = ...,
        password: Union[ConfigStr, str] = ...,
        database: str = ...,
        schema: str = ...,
        warehouse: Optional[str] = ...,
        role: Optional[str] = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    @overload
    def add_or_update_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: Union[
            ConfigStr, SnowflakeDsn, str, SnowflakeConnectionDetails, dict[str, str]
        ] = ...,
        create_temp_table: bool = ...,
        account: None = ...,
        user: None = ...,
        password: None = ...,
        database: None = ...,
        schema: None = ...,
        warehouse: None = ...,
        role: None = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    @overload
    def add_or_update_snowflake(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = ...,
        name: Optional[str] = ...,
        datasource: Optional[Datasource] = ...,
        *,
        connection_string: None = ...,
        create_temp_table: bool = ...,
        account: str = ...,
        user: str = ...,
        password: Union[ConfigStr, str] = ...,
        database: str = ...,
        schema: str = ...,
        warehouse: Optional[str] = ...,
        role: Optional[str] = ...,
        numpy: bool = ...,
    ) -> SnowflakeDatasource: ...
    def delete_snowflake(
        self,
        name: str,
    ) -> None: ...
    def add_databricks_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, DatabricksDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> DatabricksSQLDatasource: ...
    def update_databricks_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, DatabricksDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> DatabricksSQLDatasource: ...
    def add_or_update_databricks_sql(
        self,
        name_or_datasource: Optional[Union[str, Datasource]] = None,
        name: Optional[str] = None,
        datasource: Optional[Datasource] = None,
        *,
        connection_string: Union[ConfigStr, DatabricksDsn, str] = ...,
        create_temp_table: bool = True,
    ) -> DatabricksSQLDatasource: ...
    def delete_databricks_sql(
        self,
        name: str,
    ) -> None: ...
    def add_fabric_powerbi(
        self,
        name: Optional[str] = None,
        datasource: Optional[FabricPowerBIDatasource] = None,
        *,
        workspace: Optional[Union[uuid.UUID, str]] = None,
        dataset: Union[uuid.UUID, str] = ...,
    ) -> FabricPowerBIDatasource: ...
    def all(self) -> DatasourceDict: ...
    def delete(self, name: str) -> None: ...
    def get(self, name: str) -> Datasource: ...

def _iter_all_registered_types(
    include_datasource: bool = True, include_data_asset: bool = True
) -> Generator[tuple[str, Type[Datasource] | Type[DataAsset]], None, None]: ...



from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_name import MetricNameSuffix
from great_expectations.metrics.metric_results import ConditionValues, MetricResult


class ColumnValuesNonNullResult(MetricResult[ConditionValues]): ...


class ColumnValuesNonNull(ColumnMetric[ColumnValuesNonNullResult]):
    name = f"column_values.nonnull.{MetricNameSuffix.CONDITION.value}"


class ColumnValuesNonNullCountResult(MetricResult[int]): ...


class ColumnValuesNonNullCount(ColumnMetric[ColumnValuesNonNullCountResult]):
    """Count of non-null values in a column"""

    name = f"column_values.nonnull.{MetricNameSuffix.COUNT.value}"



from typing import List

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

spark_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []

databricks_deployment_patterns: List[IntegrationTestFixture] = [
    # unable to mock dbfs in CI
    # IntegrationTestFixture(
    #     name="databricks_deployment_patterns_file_python_configs",
    #     user_flow_script="docs/docusaurus/docs/snippets/databricks_deployment_patterns_file_python_configs.py",  # noqa: E501 # FIXME CoP
    #     data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
    #     backend_dependencies=[BackendDependencies.SPARK],
    # ),
]

emr_deployment_patterns = [
    IntegrationTestFixture(
        name="how_to_use_great_expectations_in_aws_emr_serverless",
        user_flow_script="tests/integration/docusaurus/deployment_patterns/aws_emr_serverless_deployment_patterns.py",
        backend_dependencies=[BackendDependencies.SPARK],
    ),
]

creating_custom_expectations = [
    IntegrationTestFixture(
        name="expect_queried_column_value_frequency_to_meet_threshold",
        user_flow_script="docs/docusaurus/docs/oss/guides/expectations/creating_custom_expectations/expect_queried_column_value_frequency_to_meet_threshold.py",
        backend_dependencies=[BackendDependencies.SPARK],
    ),
    IntegrationTestFixture(
        name="expect_queried_table_row_count_to_be",
        user_flow_script="docs/docusaurus/docs/oss/guides/expectations/creating_custom_expectations/expect_queried_table_row_count_to_be.py",
        backend_dependencies=[BackendDependencies.SPARK],
    ),
]

fluent_datasources = [
    IntegrationTestFixture(
        name="how_to_connect_to_one_or_more_files_using_spark",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_one_or_more_files_using_spark.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
        backend_dependencies=[BackendDependencies.SPARK],
    ),
]

spark_integration_tests += connecting_to_your_data
spark_integration_tests += databricks_deployment_patterns
spark_integration_tests += fluent_datasources



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

DATA = pd.DataFrame({"foo": [1, 2, 4], "bar": [1, 1, 1]})


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnPairValuesToBeInSet(
        column_A="foo", column_B="bar", value_pairs_set=[(2, 1), (1, 1)], mostly=0.5
    )

    # act
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.BASIC)

    # assert
    assert result.success

    result_dict = result.to_json_dict()["result"]
    assert type(result_dict) is dict

    # these are not deterministic
    result_dict.pop("unexpected_index_query", None)
    result_dict.pop("unexpected_index_list", None)

    assert result_dict == {
        "element_count": 3,
        "missing_count": 0,
        "missing_percent": 0.0,
        "partial_unexpected_list": [
            [4, 1],
        ],
        "unexpected_count": 1,
        "unexpected_percent": 33.33333333333333,
        "unexpected_percent_nonmissing": 33.33333333333333,
        "unexpected_percent_total": 33.33333333333333,
    }


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param([(2, 1), (1, 1)], True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_value_pairs_set_(
    batch_for_datasource: Batch, suite_param_value: list[tuple], expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_pair_values_to_be_in_set"

    expectation = gxe.ExpectColumnPairValuesToBeInSet(
        column_A="foo",
        column_B="bar",
        value_pairs_set=[(2, 1), (1, 1)],
        mostly=0.5,
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("both_values_are_missing", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_ignore_row_if_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_pair_values_to_be_in_set"

    expectation = gxe.ExpectColumnPairValuesToBeInSet(
        column_A="foo",
        column_B="bar",
        value_pairs_set=[(2, 1), (1, 1)],
        mostly=0.5,
        ignore_row_if={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



import os
import sqlite3
import typing
from logging import Logger
from typing import (
    AbstractSet,
    Any,
    Callable,
    ClassVar,
    Hashable,
    Iterable,
    List,
    Literal,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pandas as pd
from typing_extensions import TypeAlias

from great_expectations._docs_decorators import (
    deprecated_argument,
)
from great_expectations._docs_decorators import (
    public_api as public_api,
)
from great_expectations.compatibility import pydantic, sqlalchemy
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.partitioners import ColumnPartitioner
from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
from great_expectations.datasource.fluent.dynamic_pandas import (
    CompressionOptions,
    CSVEngine,
    FilePath,
    IndexLabel,
    StorageOptions,
)
from great_expectations.datasource.fluent.interfaces import (
    Batch,
    BatchMetadata,
    BatchParameters,
    BatchRequest,
    DataAsset,
    Datasource,
)
from great_expectations.execution_engine import PandasExecutionEngine

_EXCLUDE_TYPES_FROM_JSON: list[Type]

MappingIntStrAny: TypeAlias = Mapping[Union[int, str], Any]
AbstractSetIntStr: TypeAlias = AbstractSet[Union[int, str]]
logger: Logger

class PandasDatasourceError(Exception): ...

class _PandasDataAsset(DataAsset):
    _EXCLUDE_FROM_READER_OPTIONS: ClassVar[Set[str]]

    def _get_reader_method(self) -> str: ...
    @override
    def test_connection(self) -> None: ...
    def batch_parameters_template(self) -> BatchParameters: ...
    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch: ...
    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> list[dict]: ...
    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = ...,
        batch_slice: Optional[BatchSlice] = ...,
        partitioner: Optional[ColumnPartitioner] = ...,
    ) -> BatchRequest: ...
    def add_batch_definition_whole_dataframe(self, name: str) -> BatchDefinition: ...
    @override
    def _validate_batch_request(self, batch_request: BatchRequest) -> None: ...
    @override
    def json(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        skip_defaults: Union[bool, None] = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **dumps_kwargs: Any,
    ) -> str: ...

class ClipboardAsset(_PandasDataAsset): ...
class CSVAsset(_PandasDataAsset): ...
class ExcelAsset(_PandasDataAsset): ...
class FeatherAsset(_PandasDataAsset): ...
class FWFAsset(_PandasDataAsset): ...
class GBQAsset(_PandasDataAsset): ...
class HDFAsset(_PandasDataAsset): ...
class HTMLAsset(_PandasDataAsset): ...
class JSONAsset(_PandasDataAsset): ...
class ORCAsset(_PandasDataAsset): ...
class ParquetAsset(_PandasDataAsset): ...
class PickleAsset(_PandasDataAsset): ...
class SQLAsset(_PandasDataAsset): ...
class SQLQueryAsset(_PandasDataAsset): ...
class SQLTableAsset(_PandasDataAsset): ...
class SASAsset(_PandasDataAsset): ...
class SPSSAsset(_PandasDataAsset): ...
class StataAsset(_PandasDataAsset): ...
class TableAsset(_PandasDataAsset): ...
class XMLAsset(_PandasDataAsset): ...

class DataFrameAsset(_PandasDataAsset):
    type: Literal["dataframe"]

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = ...,
        batch_slice: Optional[BatchSlice] = ...,
        partitioner: Optional[ColumnPartitioner] = ...,
    ) -> BatchRequest: ...
    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch: ...
    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]: ...

_PandasDataAssetT = TypeVar("_PandasDataAssetT", bound=_PandasDataAsset)

class _PandasDatasource(Datasource):
    asset_types: ClassVar[Sequence[Type[DataAsset]]]
    assets: MutableSequence[_PandasDataAssetT]  # type: ignore[valid-type] # FIXME CoP
    @property
    @override
    def execution_engine_type(self) -> Type[PandasExecutionEngine]: ...
    @override
    def test_connection(self, test_assets: bool = ...) -> None: ...
    @override
    def json(
        self,
        *,
        include: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        exclude: Union[AbstractSetIntStr, MappingIntStrAny, None] = ...,
        by_alias: bool = ...,
        skip_defaults: Union[bool, None] = ...,
        exclude_unset: bool = ...,
        exclude_defaults: bool = ...,
        exclude_none: bool = ...,
        encoder: Union[Callable[[Any], Any], None] = ...,
        models_as_dict: bool = ...,
        **dumps_kwargs: Any,
    ) -> str: ...

_DYNAMIC_ASSET_TYPES: list[Type[_PandasDataAsset]]

class PandasDatasource(_PandasDatasource):
    asset_types: ClassVar[Sequence[Type[DataAsset]]]
    type: Literal["pandas"]
    assets: List[_PandasDataAsset]
    @override
    def test_connection(self, test_assets: bool = ...) -> None: ...
    @deprecated_argument(
        argument_name="dataframe",
        message='The "dataframe" argument is no longer part of "PandasDatasource.add_dataframe_asset()" method call; instead, "dataframe" is the required argument to "DataFrameAsset.build_batch_request()" method.',  # noqa: E501 # FIXME CoP
        version="0.16.15",
    )
    def add_dataframe_asset(
        self,
        name: str,
        *,
        dataframe: Optional[pd.DataFrame] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
    ) -> DataFrameAsset: ...
    def read_dataframe(
        self,
        dataframe: pd.DataFrame,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
    ) -> Batch: ...
    def add_clipboard_asset(
        self,
        name: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: str = "\\s+",
        kwargs: typing.Union[dict, None] = ...,
    ) -> ClipboardAsset: ...
    def add_csv_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: typing.Union[str, None] = ...,
        delimiter: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None, Literal["infer"]] = "infer",
        names: Union[Sequence[Hashable], None] = ...,
        index_col: Union[IndexLabel, Literal[False], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        prefix: str = ...,
        mangle_dupe_cols: bool = ...,
        dtype: typing.Union[dict, None] = ...,
        engine: Union[CSVEngine, None] = ...,
        converters: typing.Any = ...,
        true_values: typing.Any = ...,
        false_values: typing.Any = ...,
        skipinitialspace: bool = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        skipfooter: int = 0,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        skip_blank_lines: bool = ...,
        parse_dates: typing.Any = ...,
        infer_datetime_format: bool = ...,
        keep_date_col: bool = ...,
        date_parser: typing.Any = ...,
        dayfirst: bool = ...,
        cache_dates: bool = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        lineterminator: typing.Union[str, None] = ...,
        quotechar: str = '"',
        quoting: int = 0,
        doublequote: bool = ...,
        escapechar: typing.Union[str, None] = ...,
        comment: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        dialect: typing.Union[str, None] = ...,
        error_bad_lines: typing.Union[bool, None] = ...,
        warn_bad_lines: typing.Union[bool, None] = ...,
        on_bad_lines: typing.Any = ...,
        delim_whitespace: bool = ...,
        low_memory: typing.Any = ...,
        memory_map: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> CSVAsset: ...
    def add_excel_asset(
        self,
        name: str,
        io: os.PathLike | str | bytes,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        sheet_name: typing.Union[str, int, None] = 0,
        header: Union[int, Sequence[int], None] = 0,
        names: typing.Union[typing.List[str], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        true_values: Union[Iterable[Hashable], None] = ...,
        false_values: Union[Iterable[Hashable], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        parse_dates: typing.Union[typing.List, typing.Dict, bool] = ...,
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        comment: typing.Union[str, None] = ...,
        skipfooter: int = 0,
        convert_float: typing.Union[bool, None] = ...,
        mangle_dupe_cols: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> ExcelAsset: ...
    def add_feather_asset(
        self,
        name: str,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: Union[Sequence[Hashable], None] = ...,
        use_threads: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> FeatherAsset: ...
    def add_fwf_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        colspecs: Union[Sequence[Tuple[int, int]], str, None] = ...,
        widths: Union[Sequence[int], None] = ...,
        infer_nrows: int = ...,
        kwargs: Optional[dict] = ...,
    ) -> FWFAsset: ...
    def add_gbq_asset(
        self,
        name: str,
        query: str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        project_id: typing.Union[str, None] = ...,
        index_col: typing.Union[str, None] = ...,
        col_order: typing.Union[typing.List[str], None] = ...,
        reauth: bool = ...,
        auth_local_webserver: bool = ...,
        dialect: typing.Union[str, None] = ...,
        location: typing.Union[str, None] = ...,
        configuration: typing.Union[typing.Dict[str, typing.Any], None] = ...,
        credentials: typing.Any = ...,
        use_bqstorage_api: typing.Union[bool, None] = ...,
        max_results: typing.Union[int, None] = ...,
        progress_bar_type: typing.Union[str, None] = ...,
    ) -> GBQAsset: ...
    def add_hdf_asset(
        self,
        name: str,
        path_or_buf: str | os.PathLike | pd.HDFStore,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        key: typing.Any = ...,
        mode: str = "r",
        errors: str = "strict",
        where: typing.Union[str, typing.List, None] = ...,
        start: typing.Union[int, None] = ...,
        stop: typing.Union[int, None] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> HDFAsset: ...
    def add_html_asset(
        self,
        name: str,
        io: os.PathLike | str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        match: Union[str, typing.Pattern] = ".+",
        flavor: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        attrs: typing.Union[typing.Dict[str, str], None] = ...,
        parse_dates: bool = ...,
        thousands: typing.Union[str, None] = ",",
        encoding: typing.Union[str, None] = ...,
        decimal: str = ".",
        converters: typing.Union[typing.Dict, None] = ...,
        na_values: Union[Iterable[object], None] = ...,
        keep_default_na: bool = ...,
        displayed_only: bool = ...,
    ) -> HTMLAsset: ...
    def add_json_asset(
        self,
        name: str,
        path_or_buf: pydantic.Json | pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        orient: typing.Union[str, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        convert_axes: typing.Any = ...,
        convert_dates: typing.Union[bool, typing.List[str]] = ...,
        keep_default_dates: bool = ...,
        numpy: bool = ...,
        precise_float: bool = ...,
        date_unit: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        lines: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        nrows: typing.Union[int, None] = ...,
        storage_options: StorageOptions = ...,
    ) -> JSONAsset: ...
    def add_orc_asset(
        self,
        name: str,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> ORCAsset: ...
    def add_parquet_asset(
        self,
        name: str,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        engine: str = "auto",
        columns: typing.Union[typing.List[str], None] = ...,
        storage_options: StorageOptions = ...,
        use_nullable_dtypes: bool = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> Optional[ParquetAsset]: ...
    def add_pickle_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> PickleAsset: ...
    def add_sas_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        format: typing.Union[str, None] = ...,
        index: Union[Hashable, None] = ...,
        encoding: typing.Union[str, None] = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
    ) -> SASAsset: ...
    def add_spss_asset(
        self,
        name: str,
        path: pydantic.FilePath,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        convert_categoricals: bool = ...,
    ) -> SPSSAsset: ...
    def add_sql_asset(
        self,
        name: str,
        sql: sa.select | sa.text | str,  # type: ignore[valid-type] # FIXME CoP
        con: sqlalchemy.Engine | sqlite3.Connection | str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        params: typing.Any = ...,
        parse_dates: typing.Any = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
    ) -> SQLAsset: ...
    def add_sql_query_asset(
        self,
        name: str,
        sql: sa.select | sa.text | str,  # type: ignore[valid-type] # FIXME CoP
        con: sqlalchemy.Engine | sqlite3.Connection | str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        params: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        parse_dates: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
        dtype: typing.Union[dict, None] = ...,
    ) -> SQLQueryAsset: ...
    def add_sql_table_asset(
        self,
        name: str,
        table_name: str,
        con: sqlalchemy.Engine | str,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        schema: typing.Union[str, None] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        parse_dates: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
    ) -> SQLTableAsset: ...
    def add_stata_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        convert_dates: bool = ...,
        convert_categoricals: bool = ...,
        index_col: typing.Union[str, None] = ...,
        convert_missing: bool = ...,
        preserve_dtypes: bool = ...,
        columns: Union[Sequence[str], None] = ...,
        order_categoricals: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> StataAsset: ...
    def add_table_asset(
        self,
        name: str,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: typing.Union[str, None] = ...,
        delimiter: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None, Literal["infer"]] = "infer",
        names: Union[Sequence[Hashable], None] = ...,
        index_col: Union[IndexLabel, Literal[False], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        prefix: str = ...,
        mangle_dupe_cols: bool = ...,
        dtype: typing.Union[dict, None] = ...,
        engine: Union[CSVEngine, None] = ...,
        converters: typing.Any = ...,
        true_values: typing.Any = ...,
        false_values: typing.Any = ...,
        skipinitialspace: bool = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        skipfooter: int = 0,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        skip_blank_lines: bool = ...,
        parse_dates: typing.Any = ...,
        infer_datetime_format: bool = ...,
        keep_date_col: bool = ...,
        date_parser: typing.Any = ...,
        dayfirst: bool = ...,
        cache_dates: bool = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        lineterminator: typing.Union[str, None] = ...,
        quotechar: str = '"',
        quoting: int = 0,
        doublequote: bool = ...,
        escapechar: typing.Union[str, None] = ...,
        comment: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        dialect: typing.Union[str, None] = ...,
        error_bad_lines: typing.Union[bool, None] = ...,
        warn_bad_lines: typing.Union[bool, None] = ...,
        on_bad_lines: typing.Any = ...,
        delim_whitespace: typing.Any = ...,
        low_memory: typing.Any = ...,
        memory_map: bool = ...,
        float_precision: typing.Union[str, None] = ...,
        storage_options: StorageOptions = ...,
    ) -> TableAsset: ...
    def add_xml_asset(
        self,
        name: str,
        path_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        xpath: str = "./*",
        namespaces: typing.Union[typing.Dict[str, str], None] = ...,
        elems_only: bool = ...,
        attrs_only: bool = ...,
        names: Union[Sequence[str], None] = ...,
        dtype: typing.Union[dict, None] = ...,
        encoding: typing.Union[str, None] = "utf-8",
        stylesheet: Union[FilePath, None] = ...,
        iterparse: typing.Union[typing.Dict[str, typing.List[str]], None] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> XMLAsset: ...
    def read_clipboard(
        self,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: str = r"\s+",
        kwargs: typing.Union[dict, None] = ...,
    ) -> Batch: ...
    def read_csv(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: typing.Union[str, None] = ...,
        delimiter: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None, Literal["infer"]] = "infer",
        names: Union[Sequence[Hashable], None] = ...,
        index_col: Union[IndexLabel, Literal[False], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        prefix: str = ...,
        mangle_dupe_cols: bool = ...,
        dtype: typing.Union[dict, None] = ...,
        engine: Union[CSVEngine, None] = ...,
        converters: typing.Any = ...,
        true_values: typing.Any = ...,
        false_values: typing.Any = ...,
        skipinitialspace: bool = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        skipfooter: int = 0,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        skip_blank_lines: bool = ...,
        parse_dates: typing.Any = ...,
        infer_datetime_format: bool = ...,
        keep_date_col: bool = ...,
        date_parser: typing.Any = ...,
        dayfirst: bool = ...,
        cache_dates: bool = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        lineterminator: typing.Union[str, None] = ...,
        quotechar: str = '"',
        quoting: int = 0,
        doublequote: bool = ...,
        escapechar: typing.Union[str, None] = ...,
        comment: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        dialect: typing.Union[str, None] = ...,
        error_bad_lines: typing.Union[bool, None] = ...,
        warn_bad_lines: typing.Union[bool, None] = ...,
        on_bad_lines: typing.Any = ...,
        delim_whitespace: bool = ...,
        low_memory: typing.Any = ...,
        memory_map: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_excel(
        self,
        io: os.PathLike | str | bytes,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sheet_name: typing.Union[str, int, None] = 0,
        header: Union[int, Sequence[int], None] = 0,
        names: typing.Union[typing.List[str], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        true_values: Union[Iterable[Hashable], None] = ...,
        false_values: Union[Iterable[Hashable], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        parse_dates: typing.Union[typing.List, typing.Dict, bool] = ...,
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        comment: typing.Union[str, None] = ...,
        skipfooter: int = 0,
        convert_float: typing.Union[bool, None] = ...,
        mangle_dupe_cols: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_feather(
        self,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: Union[Sequence[Hashable], None] = ...,
        use_threads: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_fwf(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        batch_metadata: Optional[BatchMetadata] = ...,
        colspecs: Union[Sequence[Tuple[int, int]], str, None] = ...,
        widths: Union[Sequence[int], None] = ...,
        infer_nrows: int = ...,
        kwargs: Optional[dict] = ...,
    ) -> Batch: ...
    def read_gbq(
        self,
        query: str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        project_id: typing.Union[str, None] = ...,
        index_col: typing.Union[str, None] = ...,
        col_order: typing.Union[typing.List[str], None] = ...,
        reauth: bool = ...,
        auth_local_webserver: bool = ...,
        dialect: typing.Union[str, None] = ...,
        location: typing.Union[str, None] = ...,
        configuration: typing.Union[typing.Dict[str, typing.Any], None] = ...,
        credentials: typing.Any = ...,
        use_bqstorage_api: typing.Union[bool, None] = ...,
        max_results: typing.Union[int, None] = ...,
        progress_bar_type: typing.Union[str, None] = ...,
    ) -> Batch: ...
    def read_hdf(
        self,
        path_or_buf: pd.HDFStore | os.PathLike | str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        key: typing.Any = ...,
        mode: str = "r",
        errors: str = "strict",
        where: typing.Union[str, typing.List, None] = ...,
        start: typing.Union[int, None] = ...,
        stop: typing.Union[int, None] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> Batch: ...
    def read_html(
        self,
        io: os.PathLike | str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        match: Union[str, typing.Pattern] = ".+",
        flavor: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        attrs: typing.Union[typing.Dict[str, str], None] = ...,
        parse_dates: bool = ...,
        thousands: typing.Union[str, None] = ",",
        encoding: typing.Union[str, None] = ...,
        decimal: str = ".",
        converters: typing.Union[typing.Dict, None] = ...,
        na_values: Union[Iterable[object], None] = ...,
        keep_default_na: bool = ...,
        displayed_only: bool = ...,
    ) -> Batch: ...
    def read_json(
        self,
        path_or_buf: pydantic.Json | pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        orient: typing.Union[str, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        convert_axes: typing.Any = ...,
        convert_dates: typing.Union[bool, typing.List[str]] = ...,
        keep_default_dates: bool = ...,
        numpy: bool = ...,
        precise_float: bool = ...,
        date_unit: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        lines: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        nrows: typing.Union[int, None] = ...,
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_orc(
        self,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> Batch: ...
    def read_parquet(
        self,
        path: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        engine: str = "auto",
        columns: typing.Union[typing.List[str], None] = ...,
        storage_options: StorageOptions = ...,
        use_nullable_dtypes: bool = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> Batch: ...
    def read_pickle(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_sas(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        format: typing.Union[str, None] = ...,
        index: Union[Hashable, None] = ...,
        encoding: typing.Union[str, None] = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
    ) -> Batch: ...
    def read_spss(
        self,
        path: pydantic.FilePath,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        convert_categoricals: bool = ...,
    ) -> Batch: ...
    def read_sql(
        self,
        sql: sa.select | sa.text | str,  # type: ignore[valid-type] # FIXME CoP
        con: sqlalchemy.Engine | sqlite3.Connection | str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        params: typing.Any = ...,
        parse_dates: typing.Any = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
    ) -> Batch: ...
    def read_sql_query(
        self,
        sql: sa.select | sa.text | str,  # type: ignore[valid-type] # FIXME CoP
        con: sqlalchemy.Engine | sqlite3.Connection | str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        params: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        parse_dates: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
        dtype: typing.Union[dict, None] = ...,
    ) -> Batch: ...
    def read_sql_table(
        self,
        table_name: str,
        con: sqlalchemy.Engine | str,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        schema: typing.Union[str, None] = ...,
        index_col: typing.Union[str, typing.List[str], None] = ...,
        coerce_float: bool = ...,
        parse_dates: typing.Union[typing.List[str], typing.Dict[str, str], None] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        chunksize: typing.Union[int, None] = ...,
    ) -> Batch: ...
    def read_stata(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        convert_dates: bool = ...,
        convert_categoricals: bool = ...,
        index_col: typing.Union[str, None] = ...,
        convert_missing: bool = ...,
        preserve_dtypes: bool = ...,
        columns: Union[Sequence[str], None] = ...,
        order_categoricals: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_table(
        self,
        filepath_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: typing.Union[str, None] = ...,
        delimiter: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None, Literal["infer"]] = "infer",
        names: Union[Sequence[Hashable], None] = ...,
        index_col: Union[IndexLabel, Literal[False], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        prefix: str = ...,
        mangle_dupe_cols: bool = ...,
        dtype: typing.Union[dict, None] = ...,
        engine: Union[CSVEngine, None] = ...,
        converters: typing.Any = ...,
        true_values: typing.Any = ...,
        false_values: typing.Any = ...,
        skipinitialspace: bool = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        skipfooter: int = 0,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        skip_blank_lines: bool = ...,
        parse_dates: typing.Any = ...,
        infer_datetime_format: bool = ...,
        keep_date_col: bool = ...,
        date_parser: typing.Any = ...,
        dayfirst: bool = ...,
        cache_dates: bool = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        lineterminator: typing.Union[str, None] = ...,
        quotechar: str = '"',
        quoting: int = 0,
        doublequote: bool = ...,
        escapechar: typing.Union[str, None] = ...,
        comment: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        dialect: typing.Union[str, None] = ...,
        error_bad_lines: typing.Union[bool, None] = ...,
        warn_bad_lines: typing.Union[bool, None] = ...,
        on_bad_lines: typing.Any = ...,
        delim_whitespace: typing.Any = ...,
        low_memory: typing.Any = ...,
        memory_map: bool = ...,
        float_precision: typing.Union[str, None] = ...,
        storage_options: StorageOptions = ...,
    ) -> Batch: ...
    def read_xml(
        self,
        path_or_buffer: pydantic.FilePath | pydantic.AnyUrl,
        *,
        asset_name: Optional[str] = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        xpath: str = "./*",
        namespaces: typing.Union[typing.Dict[str, str], None] = ...,
        elems_only: bool = ...,
        attrs_only: bool = ...,
        names: Union[Sequence[str], None] = ...,
        dtype: typing.Union[dict, None] = ...,
        encoding: typing.Union[str, None] = "utf-8",
        stylesheet: Union[FilePath, None] = ...,
        iterparse: typing.Union[typing.Dict[str, typing.List[str]], None] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> Batch: ...



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Literal, Type, Union

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.pydantic import (
    AnyUrl,
    BaseModel,
    validator,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource
from great_expectations.execution_engine.redshift_execution_engine import RedshiftExecutionEngine

if TYPE_CHECKING:
    from great_expectations.execution_engine.sqlalchemy_execution_engine import (
        SqlAlchemyExecutionEngine,
    )


class RedshiftDsn(AnyUrl):
    allowed_schemes = {
        "redshift+psycopg2",
    }


@public_api
class RedshiftSSLModes(Enum):
    """
    Enum for the different SSL modes supported by the Redshift database.
    """

    DISABLE = "disable"
    ALLOW = "allow"
    PREFER = "prefer"
    REQUIRE = "require"
    VERIFY_CA = "verify-ca"
    VERIFY_FULL = "verify-full"


@public_api
class RedshiftConnectionDetails(BaseModel):
    """
    Information needed to connect to a Redshift database.
    """

    user: str
    password: Union[ConfigStr, str]
    host: str
    port: int
    database: str
    sslmode: RedshiftSSLModes


@public_api
class RedshiftDatasource(SQLDatasource):
    """Adds a Redshift datasource to the data context using psycopg2.

    Args:
        name: The name of this Redshift datasource.
        connection_string: The SQLAlchemy connection string used to connect to the Redshift database
            For example:
            "redshift+psycopg2://user:password@host.amazonaws.com:5439/database?sslmode=sslmode".
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose
            values are TableAsset or QueryAsset objects.
    """

    type: Literal["redshift"] = "redshift"  # type: ignore[assignment] # This is a hardcoded constant
    connection_string: Union[ConfigStr, dict, RedshiftConnectionDetails, RedshiftDsn]  # type: ignore[assignment] # Deviation from parent class as individual args are supported for connection

    @validator("connection_string", pre=True)
    def _build_connection_string_from_connection_details(
        cls, connection_string: dict | RedshiftConnectionDetails | str
    ) -> str:
        """
        If a dict or RedshiftConnectionDetails object is provided, construct the formatted
        connection string.
        """
        if isinstance(connection_string, str):
            return connection_string

        if isinstance(connection_string, dict):
            connection_details = RedshiftConnectionDetails(**connection_string)
        elif isinstance(connection_string, RedshiftConnectionDetails):
            connection_details = connection_string
        else:
            raise TypeError("Invalid connection_string type: ", type(connection_string))  # noqa: TRY003
        connection_string = f"redshift+psycopg2://{connection_details.user}:{connection_details.password}@{connection_details.host}:{connection_details.port}/{connection_details.database}?sslmode={connection_details.sslmode.value}"
        return connection_string

    @property
    @override
    def execution_engine_type(self) -> Type[SqlAlchemyExecutionEngine]:
        """Returns the default execution engine type."""
        return RedshiftExecutionEngine



from typing import Any

from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnSampleValuesResult(MetricResult[list[Any]]): ...


class ColumnSampleValues(ColumnMetric[ColumnSampleValuesResult]):
    """
    This metric returns a list of sample values from the column.

    It is only supported for SQLAlchemy execution engines at this time.

    Args:
        count: The number of sample values to return.
    """

    name = "column.sample_values"

    count: int = 20



import pandas as pd

from great_expectations.metrics.batch.batch import BatchMetric
from great_expectations.metrics.metric_results import MetricResult


class SampleValuesResult(MetricResult[pd.DataFrame]): ...


class SampleValues(BatchMetric[SampleValuesResult]):
    """Sample rows from a table"""

    name = "table.head"
    n_rows: int = 10



from __future__ import annotations

from typing import (
    ClassVar,
    List,
    Sequence,
    Type,
)

from great_expectations.datasource.fluent import _SparkDatasource
from great_expectations.datasource.fluent.data_asset.path.spark.spark_asset import (
    SPARK_PATH_ASSET_TYPES,
    SPARK_PATH_ASSET_UNION,
)
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,  # noqa: TC001  # pydantic requires this type at runtime
)


class _SparkFilePathDatasource(_SparkDatasource):
    # class attributes
    asset_types: ClassVar[Sequence[Type[DataAsset]]] = SPARK_PATH_ASSET_TYPES

    # instance attributes
    assets: List[SPARK_PATH_ASSET_UNION] = []



"""Tests to ensure core validation flows work with filesystem assets

NOTE: assertions here take the form of asserting that expectations pass
based on knowledge of the data in the test set.

Suites also assert that we only get the expected number of rows (they should all have 10)
"""

import re
from typing import Dict, Optional

import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    CSVAsset as PandasCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import (
    CSVAsset as SparkCSVAsset,
)

DATASOURCE_NAME = "file system"
ASSET_NAME = "first ten trips in each file"

# constants for what we know about the test data
BATCHING_REGEX = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
COLUMN_NAME = "passenger_count"
ALL_VALUES = list(range(1, 11))
VALUES_FOR_OLDEST_DATE = [1, 2, 3, 6]
VALUES_FOR_MOST_RECENT_DATE = [1, 2, 3, 4]
MY_FAVORITE_MONTH = {"year": "2020", "month": "09"}
VALUES_ON_MY_FAVORITE_MONTH = [0, 1]


@pytest.fixture
def context() -> AbstractDataContext:
    return gx.get_context(mode="ephemeral")


@pytest.fixture
def expect_10_rows():
    return gxe.ExpectTableRowCountToEqual(value=10)


@pytest.fixture
def pandas_file_system_asset(context: AbstractDataContext) -> PandasCSVAsset:
    datasource = context.data_sources.add_pandas_filesystem(
        DATASOURCE_NAME,
        base_directory="tests/test_sets/taxi_yellow_tripdata_samples/first_ten_trips_in_each_file",  # type: ignore [arg-type]
    )
    data_asset = datasource.add_csv_asset(name=ASSET_NAME)

    return data_asset


@pytest.fixture
def pandas_filesystem_whole_table_batch_definition(
    pandas_file_system_asset: PandasCSVAsset,
) -> BatchDefinition:
    return pandas_file_system_asset.add_batch_definition("no batching regex")


@pytest.fixture
def pandas_filesystem_monthly_batch_definition(
    pandas_file_system_asset: PandasCSVAsset,
) -> BatchDefinition:
    return pandas_file_system_asset.add_batch_definition_monthly(  # type: ignore[attr-defined] # FIXME CoP
        "monthly",
        re.compile(BATCHING_REGEX),
    )


@pytest.fixture
def pandas_filesystem_monthly_batch_definition_descending(
    pandas_file_system_asset: PandasCSVAsset,
) -> BatchDefinition:
    return pandas_file_system_asset.add_batch_definition_monthly(  # type: ignore[attr-defined] # FIXME CoP
        "monthly",
        re.compile(BATCHING_REGEX),
        sort_ascending=False,
    )


@pytest.fixture
def spark_file_system_asset(context: AbstractDataContext) -> SparkCSVAsset:
    datasource = context.data_sources.add_spark_filesystem(
        DATASOURCE_NAME,
        base_directory="tests/test_sets/taxi_yellow_tripdata_samples/first_ten_trips_in_each_file",  # type: ignore [arg-type]
    )
    data_asset = datasource.add_csv_asset(
        name=ASSET_NAME,
        header=True,
        infer_schema=True,
    )

    return data_asset


@pytest.fixture
def spark_filesystem_whole_table_batch_definition(
    spark_file_system_asset: SparkCSVAsset,
) -> BatchDefinition:
    return spark_file_system_asset.add_batch_definition("no batching regex")


@pytest.fixture
def spark_filesystem_monthly_batch_definition(
    spark_file_system_asset: SparkCSVAsset,
) -> BatchDefinition:
    return spark_file_system_asset.add_batch_definition_monthly(
        "monthly",
        re.compile(BATCHING_REGEX),
    )


@pytest.fixture
def spark_filesystem_monthly_batch_definition_descending(
    spark_file_system_asset: SparkCSVAsset,
) -> BatchDefinition:
    return spark_file_system_asset.add_batch_definition_monthly(
        "monthly",
        re.compile(BATCHING_REGEX),
        sort_ascending=False,
    )


def _create_test_cases():
    """Create our test cases.

    With each flow, we want to see that we can validate an entire asset,
    as well as subsets of the asset, including sorting and using batch parameters.

    The positional arguments are:
    - An Expectation
    - The fixture name for the batch definition
    - The batch_parameters to pass in during validation
    """
    test_cases = []

    for datasource_type, mark in [("pandas", pytest.mark.filesystem), ("spark", pytest.mark.spark)]:
        test_cases.extend(
            [
                pytest.param(
                    gxe.ExpectColumnDistinctValuesToEqualSet(
                        column=COLUMN_NAME, value_set=VALUES_FOR_MOST_RECENT_DATE
                    ),
                    f"{datasource_type}_filesystem_whole_table_batch_definition",
                    None,  # no batch parameters
                    id=f"{datasource_type}: no batching regex - takes the last file",
                    marks=[mark],
                ),
                pytest.param(
                    gxe.ExpectColumnDistinctValuesToEqualSet(
                        column=COLUMN_NAME, value_set=VALUES_FOR_MOST_RECENT_DATE
                    ),
                    f"{datasource_type}_filesystem_monthly_batch_definition",
                    None,  # no batch parameters
                    id=f"{datasource_type}: ascending",
                    marks=[mark],
                ),
                pytest.param(
                    gxe.ExpectColumnDistinctValuesToEqualSet(
                        column=COLUMN_NAME, value_set=VALUES_FOR_OLDEST_DATE
                    ),
                    f"{datasource_type}_filesystem_monthly_batch_definition_descending",
                    None,  # no batch parameters
                    id=f"{datasource_type}: descending",
                    marks=[mark],
                ),
                pytest.param(
                    gxe.ExpectColumnDistinctValuesToEqualSet(
                        column=COLUMN_NAME, value_set=VALUES_ON_MY_FAVORITE_MONTH
                    ),
                    f"{datasource_type}_filesystem_monthly_batch_definition",
                    MY_FAVORITE_MONTH,
                    id=f"{datasource_type}: batch params",
                    marks=[mark],
                ),
            ]
        )
    return test_cases


@pytest.mark.parametrize(
    "batch_definition_fixture_name",
    [
        pytest.param("pandas_filesystem_monthly_batch_definition", marks=[pytest.mark.filesystem]),
        pytest.param("spark_filesystem_monthly_batch_definition", marks=[pytest.mark.spark]),
    ],
)
def test_get_batch_identifiers_list__simple(
    batch_definition_fixture_name: str,
    request: pytest.FixtureRequest,
) -> None:
    batch_definition: BatchDefinition = request.getfixturevalue(batch_definition_fixture_name)
    batch_identifiers_list = batch_definition.get_batch_identifiers_list()

    assert len(batch_identifiers_list) == 36
    # just spot check the edges
    assert batch_identifiers_list[0] == {
        "path": "yellow_tripdata_sample_2018-01.csv",
        "year": "2018",
        "month": "01",
    }
    assert batch_identifiers_list[-1] == {
        "path": "yellow_tripdata_sample_2020-12.csv",
        "year": "2020",
        "month": "12",
    }


@pytest.mark.parametrize(
    "batch_definition_fixture_name",
    [
        pytest.param(
            "pandas_filesystem_monthly_batch_definition_descending", marks=[pytest.mark.filesystem]
        ),
        pytest.param(
            "spark_filesystem_monthly_batch_definition_descending", marks=[pytest.mark.spark]
        ),
    ],
)
def test_get_batch_identifiers_list__respects_order(
    batch_definition_fixture_name: str,
    request: pytest.FixtureRequest,
) -> None:
    batch_definition: BatchDefinition = request.getfixturevalue(batch_definition_fixture_name)
    batch_identifiers_list = batch_definition.get_batch_identifiers_list()

    assert len(batch_identifiers_list) == 36
    # just spot check the edges
    assert batch_identifiers_list[0] == {
        "path": "yellow_tripdata_sample_2020-12.csv",
        "year": "2020",
        "month": "12",
    }
    assert batch_identifiers_list[-1] == {
        "path": "yellow_tripdata_sample_2018-01.csv",
        "year": "2018",
        "month": "01",
    }


@pytest.mark.parametrize(
    "batch_definition_fixture_name",
    [
        pytest.param("pandas_filesystem_monthly_batch_definition", marks=[pytest.mark.filesystem]),
        pytest.param("spark_filesystem_monthly_batch_definition", marks=[pytest.mark.spark]),
    ],
)
def test_get_batch_identifiers_list__respects_batch_params(
    batch_definition_fixture_name: str,
    request: pytest.FixtureRequest,
) -> None:
    batch_definition: BatchDefinition = request.getfixturevalue(batch_definition_fixture_name)
    batch_identifiers_list = batch_definition.get_batch_identifiers_list(
        batch_parameters={"year": "2020"}
    )

    assert len(batch_identifiers_list) == 12
    # just spot check the edges
    assert batch_identifiers_list[0] == {
        "path": "yellow_tripdata_sample_2020-01.csv",
        "year": "2020",
        "month": "01",
    }
    assert batch_identifiers_list[-1] == {
        "path": "yellow_tripdata_sample_2020-12.csv",
        "year": "2020",
        "month": "12",
    }


@pytest.mark.parametrize(
    "batch_definition_fixture_name",
    [
        pytest.param("pandas_filesystem_monthly_batch_definition", marks=[pytest.mark.filesystem]),
        pytest.param("spark_filesystem_monthly_batch_definition", marks=[pytest.mark.spark]),
    ],
)
def test_get_batch_identifiers_list__no_batches(
    batch_definition_fixture_name: str,
    request: pytest.FixtureRequest,
) -> None:
    batch_definition: BatchDefinition = request.getfixturevalue(batch_definition_fixture_name)
    batch_identifiers_list = batch_definition.get_batch_identifiers_list(
        batch_parameters={"year": "1999"}
    )

    assert batch_identifiers_list == []


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_batch_validate_expectation(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    request: pytest.FixtureRequest,
) -> None:
    """Ensure Batch::validate(Epectation) works"""
    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    batch = batch_definition.get_batch(batch_parameters=batch_parameters)

    result = batch.validate(expectation)
    row_count_result = batch.validate(gxe.ExpectTableRowCountToEqual(value=10))

    assert row_count_result.success
    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_batch_validate_expectation_suite(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    request: pytest.FixtureRequest,
    expect_10_rows: gxe.Expectation,
) -> None:
    """Ensure Batch::validate(EpectationSuite) works"""

    suite = ExpectationSuite("my suite", expectations=[expectation, expect_10_rows])
    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    batch = batch_definition.get_batch(batch_parameters=batch_parameters)
    result = batch.validate(suite)

    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_validation_definition_run(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    context: AbstractDataContext,
    request: pytest.FixtureRequest,
    expect_10_rows: gxe.Expectation,
) -> None:
    """Ensure ValidationDefinition::run works"""

    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    suite = context.suites.add(
        ExpectationSuite("my suite", expectations=[expectation, expect_10_rows])
    )
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="whatever", data=batch_definition, suite=suite)
    )
    result = validation_definition.run(batch_parameters=batch_parameters)

    assert result.success


@pytest.mark.parametrize(
    ("expectation", "batch_definition_fixture_name", "batch_parameters"),
    _create_test_cases(),
)
def test_checkpoint_run(
    expectation: gxe.Expectation,
    batch_definition_fixture_name: str,
    batch_parameters: Optional[Dict],
    context: AbstractDataContext,
    request: pytest.FixtureRequest,
    expect_10_rows: gxe.Expectation,
) -> None:
    """Ensure Checkpoint::run works"""

    batch_definition = request.getfixturevalue(batch_definition_fixture_name)
    suite = context.suites.add(
        ExpectationSuite("my suite", expectations=[expectation, expect_10_rows])
    )
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="whatever", data=batch_definition, suite=suite)
    )
    checkpoint = context.checkpoints.add(
        Checkpoint(name="whatever", validation_definitions=[validation_definition])
    )
    result = checkpoint.run(batch_parameters=batch_parameters)

    assert result.success



from __future__ import annotations

import logging
from string import Template
from typing import ClassVar, Type

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.not_imported import is_version_greater_or_equal
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.data_context_key import StringKey
from great_expectations.data_context.store.store import Store
from great_expectations.util import filter_properties_dict

if sa:
    if is_version_greater_or_equal(sa.__version__, "1.4.0"):
        url_create_fn = sqlalchemy.URL.create
    else:
        url_create_fn = sqlalchemy.URL  # type: ignore[assignment] # FIXME CoP


logger = logging.getLogger(__name__)


class SqlAlchemyQueryStore(Store):
    """SqlAlchemyQueryStore stores queries by name, and makes it possible to retrieve the resulting value by query
    name."""  # noqa: E501 # FIXME CoP

    _key_class: ClassVar[Type] = StringKey

    def __init__(
        self,
        credentials,
        queries=None,
        store_backend=None,
        runtime_environment=None,
        store_name=None,
    ) -> None:
        if not sa:
            raise gx_exceptions.DataContextError(  # noqa: TRY003 # FIXME CoP
                "sqlalchemy module not found, but is required for SqlAlchemyQueryStore"
            )
        super().__init__(
            store_backend=store_backend,
            runtime_environment=runtime_environment,
            store_name=store_name,
        )
        if queries:
            # If queries are defined in configuration, then we load them into an InMemoryStoreBackend  # noqa: E501 # FIXME CoP
            try:
                assert isinstance(queries, dict), (
                    "SqlAlchemyQueryStore queries must be defined as a dictionary"
                )
                assert (
                    store_backend is None or store_backend["class_name"] == "InMemoryStoreBackend"
                ), (
                    "If queries are provided in configuration, then store_backend must be empty or an "  # noqa: E501 # FIXME CoP
                    "InMemoryStoreBackend"
                )
                for k, v in queries.items():
                    self._store_backend.set(tuple([k]), v)

            except (AssertionError, KeyError) as e:
                raise gx_exceptions.InvalidConfigError(str(e))

        if "engine" in credentials:
            self.engine = credentials["engine"]
        elif "url" in credentials:
            self.engine = sa.create_engine(credentials["url"])
        elif "connection_string" in credentials:
            self.engine = sa.create_engine(credentials["connection_string"])
        else:
            drivername = credentials.pop("drivername")
            options = url_create_fn(drivername, **credentials)
            self.engine = sa.create_engine(options)

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "credentials": credentials,
            "queries": queries,
            "store_backend": store_backend,
            "runtime_environment": runtime_environment,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def _convert_key(self, key):
        if isinstance(key, str):
            return StringKey(key)
        return key

    def get(self, key):  # type: ignore[explicit-override] # FIXME
        return super().get(self._convert_key(key))

    def set(self, key, value):  # type: ignore[explicit-override] # FIXME
        return super().set(self._convert_key(key), value)

    def get_query_result(self, key, query_parameters=None):
        if query_parameters is None:
            query_parameters = {}
        result = self._store_backend.get(self._convert_key(key).to_tuple())
        if isinstance(result, dict):
            query = result.get("query")
            return_type = result.get("return_type", "list")
            if return_type not in ["list", "scalar"]:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "The return_type of a SqlAlchemyQueryStore query must be one of either 'list' "
                    "or 'scalar'"
                )
        else:
            query = result
            return_type = None

        assert query, "Query must be specified to use SqlAlchemyQueryStore"

        query = Template(query).safe_substitute(query_parameters)
        with self.engine.begin() as connection:
            res = connection.execute(sa.text(query)).fetchall()
            # NOTE: 20200617 - JPC: this approach is probably overly opinionated, but we can
            # adjust based on specific user requests
            res = [val for row in res for val in row]
            if return_type == "scalar":
                [res] = res
            return res

    @property
    @override
    def config(self) -> dict:
        return self._config



from .components import (
    AtomicDiagnosticRendererType,
    AtomicPrescriptiveRendererType,
    AtomicRendererType,
    CollapseContent,
    LegacyDescriptiveRendererType,
    LegacyDiagnosticRendererType,
    LegacyPrescriptiveRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedAtomicContentSchema,
    RenderedAtomicValue,
    RenderedAtomicValueGraph,
    RenderedAtomicValueSchema,
    RenderedBootstrapTableContent,
    RenderedBulletListContent,
    RenderedComponentContent,
    RenderedContent,
    RenderedContentBlockContainer,
    RenderedDocumentContent,
    RenderedGraphContent,
    RenderedHeaderContent,
    RenderedMarkdownContent,
    RenderedSectionContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    RenderedTabsContent,
    RendererPrefix,
    TextContent,
    ValueListContent,
)
from .view import DefaultJinjaPageView

renderedAtomicValueSchema = RenderedAtomicValueSchema()
